@@ -21,6 +21,8 @@
 #include <config.h>
 #include "intprops.h"
 #include "search.h"
+#include "die.h"
+#include <error.h>
 
 struct localeinfo localeinfo;
 
@@ -53,11 +55,7 @@ static bool begline;
 void
 dfaerror (char const *mesg)
 {
-  error (EXIT_TROUBLE, 0, "%s", mesg);
-
-  /* notreached */
-  /* Tell static analyzers that this function does not return.  */
-  abort ();
+  die (EXIT_TROUBLE, 0, "%s", mesg);
 }
 
 /* For now, the sole dfawarn-eliciting condition (use of a regexp
@@ -218,7 +216,7 @@ GEAcompile (char const *pattern, size_t size, reg_syntax_t syntax_bits)
 }
 
 size_t
-EGexecute (char *buf, size_t size, size_t *match_size,
+EGexecute (char const *buf, size_t size, size_t *match_size,
            char const *start_ptr)
 {
   char const *buflim, *beg, *end, *ptr, *match, *best_match, *mb_start;
